"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class DependsOnObsolete(CloudFormationLintRule):
    """Check unneeded DepensOn Resource Configuration"""

    id = "W3005"
    shortdesc = "Check obsolete DependsOn configuration for Resources"
    description = (
        "Check if DependsOn is specified if not needed. "
        "A Ref or a Fn::GetAtt already is an implicit dependency."
    )
    source_url = (
        "https://aws.amazon.com/blogs/devops/optimize-aws-cloudformation-templates/"
    )
    tags = ["resources", "dependson", "ref", "getatt"]

    def get_resource_references(self, cfn, ref_function, resource):
        """Get tree of all resource references of a resource"""
        trees = cfn.search_deep_keys(ref_function)

        # Filter only resoureces
        trees = filter(lambda x: x[0] == "Resources", trees)
        # Filter on the given resource only
        trees = filter(lambda x: x[1] == resource, trees)

        return trees

    def check_depends_on(self, cfn, resource, key, path):
        """Check if the DependsOn is already specified"""
        matches = []

        # Get references
        trees = self.get_resource_references(cfn, "Ref", resource)

        for tree in trees:
            if tree[-1] == key:
                message = 'Obsolete DependsOn on resource ({0}), dependency already enforced by a "Ref" at {1}'
                matches.append(
                    RuleMatch(path, message.format(key, "/".join(map(str, tree[:-1]))))
                )

        # Get the GetAtt
        trees = self.get_resource_references(cfn, "Fn::GetAtt", resource)

        for tree in trees:
            # GettAtt formation is "resource : Attribute", just check the resource
            if tree[-1][0] == key:
                message = 'Obsolete DependsOn on resource ({0}), dependency already enforced by a "Fn:GetAtt" at {1}'
                matches.append(
                    RuleMatch(path, message.format(key, "/".join(map(str, tree[:-1]))))
                )

        return matches

    def match(self, cfn):
        matches = []

        resources = cfn.get_resources()

        for resource_name, resource_values in resources.items():
            depends_ons = resource_values.get("DependsOn")
            if depends_ons:
                path = ["Resources", resource_name, "DependsOn"]
                self.logger.debug("Validating unneeded DependsOn for %s", resource_name)
                if isinstance(depends_ons, list):
                    for index, depends_on in enumerate(depends_ons):
                        matches.extend(
                            self.check_depends_on(
                                cfn, resource_name, depends_on, path[:] + [index]
                            )
                        )
                else:
                    matches.extend(
                        self.check_depends_on(cfn, resource_name, depends_ons, path)
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class RecordSetName(CloudFormationLintRule):
    """Check if a Route53 Resoruce Records Name is valid with a HostedZoneName"""

    id = "E3041"
    shortdesc = "RecordSet HostedZoneName is a superdomain of Name"
    description = "In a RecordSet, the HostedZoneName must be a superdomain of the Name being validated"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-route53-recordset.html#cfn-route53-recordset-name"
    tags = ["resource", "properties", "route53"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types = ["AWS::Route53::RecordSet"]

    def match_resource_properties(self, properties, _, path, cfn):
        matches = []

        property_sets = cfn.get_object_without_conditions(
            properties, ["Name", "HostedZoneName"]
        )
        for property_set in property_sets:
            props = property_set.get("Object")
            scenario = property_set.get("Scenario")
            name = props.get("Name", None)
            hz_name = props.get("HostedZoneName", None)
            if isinstance(name, str) and isinstance(hz_name, str):
                if hz_name[-1] != ".":
                    message = "HostedZoneName must end in a dot at {}"
                    if scenario is None:
                        matches.append(
                            RuleMatch(
                                path[:] + ["HostedZoneName"],
                                message.format("/".join(map(str, path))),
                            )
                        )
                    else:
                        scenario_text = " and ".join(
                            [
                                f'when condition "{k}" is {v}'
                                for (k, v) in scenario.items()
                            ]
                        )
                        matches.append(
                            RuleMatch(
                                path[:] + ["HostedZoneName"],
                                message.format(
                                    "/".join(map(str, path)) + " " + scenario_text
                                ),
                            )
                        )
                if hz_name[-1] == ".":
                    hz_name = hz_name[:-1]
                if name[-1] == ".":
                    name = name[:-1]

                if hz_name not in [name, name[-len(hz_name) :]]:
                    message = "Name must be a superdomain of HostedZoneName at {}"
                    if scenario is None:
                        matches.append(
                            RuleMatch(
                                path[:] + ["Name"],
                                message.format("/".join(map(str, path))),
                            )
                        )
                    else:
                        scenario_text = " and ".join(
                            [
                                f'when condition "{k}" is {v}'
                                for (k, v) in scenario.items()
                            ]
                        )
                        matches.append(
                            RuleMatch(
                                path[:] + ["Name"],
                                message.format(
                                    "/".join(map(str, path)) + " " + scenario_text
                                ),
                            )
                        )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class AvailabilityZone(CloudFormationLintRule):
    """Check Availibility Zone parameter checks"""

    id = "W3010"
    shortdesc = "Availability Zone Parameters should not be hardcoded"
    description = "Check if an Availability Zone property is hardcoded."
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["parameters", "availabilityzone"]

    def __init__(self):
        """Init"""
        super().__init__()
        resource_type_specs = [
            "AWS::AutoScaling::AutoScalingGroup",
            "AWS::DAX::Cluster",
            "AWS::DMS::ReplicationInstance",
            "AWS::EC2::Host",
            "AWS::EC2::Instance",
            "AWS::EC2::Subnet",
            "AWS::EC2::Volume",
            "AWS::ElasticLoadBalancing::LoadBalancer",
            "AWS::OpsWorks::Instance",
            "AWS::RDS::DBCluster",
            "AWS::RDS::DBInstance",
        ]

        property_type_specs = [
            # Singular
            "AWS::EC2::LaunchTemplate.Placement",
            "AWS::EC2::SpotFleet.LaunchTemplateOverrides",
            "AWS::EC2::SpotFleet.SpotPlacement",
            "AWS::EMR::Cluster.PlacementType",
            "AWS::ElasticLoadBalancingV2::TargetGroup.TargetDescription",
            "AWS::Glue::Connection.PhysicalConnectionRequirements",
        ]

        for resource_type_spec in resource_type_specs:
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in property_type_specs:
            self.resource_sub_property_types.append(property_type_spec)

    # pylint: disable=W0613
    def check_az_value(self, value, path):
        matches = []

        # value of `all` is a valide exception in AWS::ElasticLoadBalancingV2::TargetGroup
        if value not in ["all"]:
            if path[-1] != ["Fn::GetAZs"]:
                message = "Don't hardcode {0} for AvailabilityZones"
                matches.append(RuleMatch(path, message.format(value)))

        return matches

    def check(self, properties, resource_type, path, cfn):
        """Check itself"""
        matches = []

        # Skip rule if CDK
        if cfn.is_cdk_template():
            return matches

        matches.extend(
            cfn.check_value(
                properties,
                "AvailabilityZone",
                path,
                check_value=self.check_az_value,
                check_ref=None,
                check_find_in_map=None,
                check_split=None,
                check_join=None,
            )
        )
        matches.extend(
            cfn.check_value(
                properties,
                "AvailabilityZones",
                path,
                check_value=self.check_az_value,
                check_ref=None,
                check_find_in_map=None,
                check_split=None,
                check_join=None,
            )
        )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        matches.extend(self.check(properties, property_type, path, cfn))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        matches.extend(self.check(properties, resource_type, path, cfn))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

import cfnlint.helpers
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class DynamicReferenceSecureString(CloudFormationLintRule):
    """Check if Dynamic Reference Secure Strings are only used in the correct locations"""

    id = "E1027"
    shortdesc = "Check dynamic references secure strings are in supported locations"
    description = (
        "Dynamic References Secure Strings are only supported for a small set of resource properties.  "
        "Validate that they are being used in the correct location when checking values "
        "and Fn::Sub in resource properties. Currently doesn't check outputs, maps, conditions, "
        "parameters, and descriptions."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/dynamic-references.html"
    tags = ["functions", "dynamic reference"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.property_specs = []
        self.resource_specs = []
        self.exceptions = {
            "AWS::DirectoryService::MicrosoftAD": "Password",
            "AWS::DirectoryService::SimpleAD": "Password",
            "AWS::ElastiCache::ReplicationGroup": "AuthToken",
            "AWS::IAM::User.LoginProfile": "Password",
            "AWS::KinesisFirehose::DeliveryStream.RedshiftDestinationConfiguration": "Password",
            "AWS::OpsWorks::App.Source": "Password",
            "AWS::OpsWorks::Stack.RdsDbInstance": "DbPassword",
            "AWS::OpsWorks::Stack.Source": "Password",
            "AWS::RDS::DBCluster": "MasterUserPassword",
            "AWS::RDS::DBInstance": "MasterUserPassword",
            "AWS::Redshift::Cluster": "MasterUserPassword",
        }

    def initialize(self, cfn):
        """Init"""
        specs = cfnlint.helpers.RESOURCE_SPECS.get(cfn.regions[0])
        self.property_specs = specs.get("PropertyTypes")
        self.resource_specs = specs.get("ResourceTypes")
        for resource_spec in self.resource_specs:
            self.resource_property_types.append(resource_spec)
        for property_spec in self.property_specs:
            self.resource_sub_property_types.append(property_spec)

    def check_dyn_ref_value(self, value, path):
        """Chec item type"""
        matches = []

        if isinstance(value, str):
            if re.match(cfnlint.helpers.REGEX_DYN_REF_SSM_SECURE, value):
                message = f'Dynamic Reference secure strings are not supported for this property at {"/".join(map(str, path[:]))}'
                matches.append(RuleMatch(path[:], message))

        return matches

    def check_value(self, value, path, **kwargs):
        """Check Value"""
        matches = []
        item_type = kwargs.get("item_type", {})
        if item_type in ["Map"]:
            if isinstance(value, dict):
                for map_key, map_value in value.items():
                    if not isinstance(map_value, dict):
                        matches.extend(
                            self.check_dyn_ref_value(map_value, path[:] + [map_key])
                        )
        else:
            matches.extend(self.check_dyn_ref_value(value, path[:]))

        return matches

    # pylint: disable=W0613
    # Need to disable for the function to work
    def check_sub(self, value, path, **kwargs):
        """Check Sub Function Dynamic References"""
        matches = []
        if isinstance(value, list):
            if isinstance(value[0], str):
                matches.extend(self.check_dyn_ref_value(value[0], path[:] + [0]))
        else:
            matches.extend(self.check_dyn_ref_value(value, path[:]))

        return matches

    def check(self, cfn, properties, specs, property_type, path):
        """Check itself"""
        matches = []

        for prop in properties:
            if prop in specs:
                if property_type in self.exceptions:
                    if prop == self.exceptions.get(property_type):
                        continue
                primitive_type = specs.get(prop).get("PrimitiveType")
                if not primitive_type:
                    primitive_type = specs.get(prop).get("PrimitiveItemType")
                if specs.get(prop).get("Type") in ["List", "Map"]:
                    item_type = specs.get(prop).get("Type")
                else:
                    item_type = None
                if primitive_type:
                    matches.extend(
                        cfn.check_value(
                            properties,
                            prop,
                            path[:],
                            check_value=self.check_value,
                            check_sub=self.check_sub,
                            primitive_type=primitive_type,
                            item_type=item_type,
                        )
                    )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        if self.property_specs.get(property_type, {}).get("Properties"):
            property_specs = self.property_specs.get(property_type, {}).get(
                "Properties", {}
            )
            matches.extend(
                self.check(cfn, properties, property_specs, property_type, path)
            )

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []
        resource_specs = self.resource_specs.get(resource_type, {}).get(
            "Properties", {}
        )
        matches.extend(self.check(cfn, properties, resource_specs, resource_type, path))

        return matches



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from typing import Any, Mapping

from cfnlint.conditions import Conditions
from cfnlint.graph import Graph
from cfnlint.template.transforms._language_extensions import language_extension
from cfnlint.template.transforms._protocols import Transformer
from cfnlint.template.transforms._sam import sam

LOGGER = logging.getLogger("cfnlint")


class Transform:
    def __init__(self) -> None:
        self.transforms: Mapping[str, Transformer] = {
            "AWS::Serverless-2016-10-31": sam,
            "AWS::LanguageExtensions": language_extension,
        }

    def transform(self, cfn: Any):
        """Transform logic"""
        matches = []
        transform_declaration = cfn.template.get("Transform", [])
        transform_type = (
            transform_declaration
            if isinstance(transform_declaration, list)
            else [transform_declaration]
        )

        if not transform_type:
            return matches

        cfn.transform_pre["Globals"] = cfn.template.get("Globals", {})
        for name in transform_type:
            if not isinstance(name, str):
                continue
            transform = self.transforms.get(name)
            if not transform:
                continue

            matches, template = transform(cfn)
            if matches:
                return matches
            cfn.template = template

        LOGGER.info("Transformed template: %s", cfn.template)
        cfn.graph = Graph(cfn)
        cfn.conditions = Conditions(cfn)
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class CodepipelineStages(CloudFormationLintRule):
    """Check if CodePipeline Stages are set up properly."""

    id = "E2540"
    shortdesc = "CodePipeline Stages"
    description = "See if CodePipeline stages are set correctly"
    source_url = "https://docs.aws.amazon.com/codepipeline/latest/userguide/reference-pipeline-structure.html#pipeline-requirements"
    tags = ["properties", "codepipeline"]

    def _format_error_message(self, message, scenario):
        """Format error message with scenario text"""
        if scenario:
            scenario_text = " When " + " and ".join(
                [f'condition "{k}" is {v}' for (k, v) in scenario.items()]
            )
            return message + scenario_text

        return message

    def check_stage_count(self, stages, path, scenario):
        """Check that there is minimum 2 stages."""
        matches = []

        if len(stages) < 2:
            message = f"CodePipeline has {len(stages)} stages. There must be at least two stages."
            matches.append(
                RuleMatch(path, self._format_error_message(message, scenario))
            )

        return matches

    def check_first_stage(self, stages, path, scenario):
        """Validate the first stage of a pipeline has source actions."""
        matches = []

        if len(stages) < 1:  # pylint: disable=C1801
            self.logger.debug(
                "Stages was empty. Should have been caught by generic linting."
            )
            return matches

        # pylint: disable=R1718
        first_stage = set(
            [a.get("ActionTypeId").get("Category") for a in stages[0]["Actions"]]
        )
        if first_stage and "Source" not in first_stage:
            message = (
                "The first stage of a pipeline must contain at least one source action."
            )
            matches.append(
                RuleMatch(
                    path + [0, "Name"], self._format_error_message(message, scenario)
                )
            )

        if len(first_stage) != 1:
            message = "The first stage of a pipeline must contain only source actions."
            matches.append(
                RuleMatch(
                    path + [0, "Name"], self._format_error_message(message, scenario)
                )
            )

        return matches

    def check_source_actions(self, stages, path, scenario):
        """Validate the all of the stages."""
        matches = []
        categories = set()

        if len(stages) < 1:  # pylint: disable=C1801
            self.logger.debug(
                "Stages was empty. Should have been caught by generic linting."
            )
            return matches

        for sidx, stage in enumerate(stages):
            for aidx, action in enumerate(stage.get("Actions", [])):
                action_type_id = action.get("ActionTypeId")
                categories.add(action_type_id.get("Category"))
                if sidx > 0 and action_type_id.get("Category") == "Source":
                    message = (
                        "Only the first stage of a pipeline may contain source actions."
                    )
                    matches.append(
                        RuleMatch(
                            path + [sidx, "Actions", aidx],
                            self._format_error_message(message, scenario),
                        )
                    )

        if not categories - set(["Source"]):
            message = "At least one stage in pipeline must contain an action that is not a source action."
            matches.append(
                RuleMatch(path, self._format_error_message(message, scenario))
            )

        return matches

    def check_names_unique(self, value, path, scenario):
        """Check that stage names are unique."""
        matches = []
        stage_names = set()
        for sidx, stage in enumerate(value):
            stage_name = stage.get("Name")
            if isinstance(stage_name, str):
                if stage_name in stage_names:
                    message = f"All stage names within a pipeline must be unique. ({stage_name})"
                    matches.append(
                        RuleMatch(
                            path + [sidx, "Name"],
                            self._format_error_message(message, scenario),
                        )
                    )
                stage_names.add(stage_name)
            else:
                self.logger.debug("Found non string for stage name: %s", stage_name)
        return matches

    def match(self, cfn):
        """Check CodePipeline stages"""
        matches = []

        resources = cfn.get_resource_properties(["AWS::CodePipeline::Pipeline"])
        for resource in resources:
            path = resource["Path"] + ["Stages"]
            properties = resource["Value"]

            s_stages = cfn.get_object_without_nested_conditions(
                properties.get("Stages"), path
            )
            for s_stage in s_stages:
                s_stage_obj = s_stage.get("Object")
                s_scenario = s_stage.get("Scenario")
                if not isinstance(s_stage_obj, list):
                    self.logger.debug(
                        "Stages not list. Should have been caught by generic linting."
                    )
                    continue

                try:
                    matches.extend(
                        self.check_stage_count(s_stage_obj, path, s_scenario)
                    )
                    matches.extend(
                        self.check_first_stage(s_stage_obj, path, s_scenario)
                    )
                    matches.extend(
                        self.check_source_actions(s_stage_obj, path, s_scenario)
                    )
                    matches.extend(
                        self.check_names_unique(s_stage_obj, path, s_scenario)
                    )
                except AttributeError as err:
                    self.logger.debug(
                        "Got AttributeError. Should have been caught by generic linting. "
                        "Ignoring the error here: %s",
                        str(err),
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.decode.node import list_node
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class AttributeMismatch(CloudFormationLintRule):
    """Check DynamoDB Attributes"""

    id = "E3039"
    shortdesc = "AttributeDefinitions / KeySchemas mismatch"
    description = (
        "Verify the set of Attributes in AttributeDefinitions and KeySchemas match"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-dynamodb-table.html"
    tags = ["resources", "dynamodb"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types = ["AWS::DynamoDB::Table"]

    def _get_key_schema_attributes(self, key_schemas_sets):
        """Get Key Schema attributes"""
        keys = set()

        for properties, _ in key_schemas_sets:
            for key in properties:
                attribute_name = key.get_safe("AttributeName", type_t=str)
                if attribute_name:
                    keys.add(key.get("AttributeName"))
        return keys

    def _get_attribute_secondary(self, property_sets):
        """Get the key schemas from secondary indexes"""
        keys = set()

        for properties, _ in property_sets:
            for index in properties:
                keys = keys.union(
                    self._get_key_schema_attributes(
                        index.get_safe("KeySchema", list_node([], None, None), [], list)
                    )
                )

        return keys

    def check_property_set(self, property_set, path):
        """Check a property set"""
        matches = []
        properties = property_set.get("Object")

        keys = set()
        attributes = set()

        for attribute in properties.get("AttributeDefinitions", []):
            attribute_name = attribute.get("AttributeName")
            if isinstance(attribute_name, str):
                attributes.add(attribute.get("AttributeName"))
            else:
                self.logger.info("attribute definitions is not using just strings")
                return matches
        keys = keys.union(
            self._get_key_schema_attributes(
                properties.get_safe("KeySchema", list_node([], None, None), [], list)
            )
        )
        keys = keys.union(
            self._get_attribute_secondary(
                properties.get_safe(
                    "GlobalSecondaryIndexes", list_node([], None, None), path, list
                )
            )
        )
        keys = keys.union(
            self._get_attribute_secondary(
                properties.get_safe(
                    "LocalSecondaryIndexes", list_node([], None, None), path, list
                )
            )
        )

        if attributes != keys:
            message = "The set of Attributes in AttributeDefinitions: {0} and KeySchemas: {1} must match at {2}"
            matches.append(
                RuleMatch(
                    path,
                    message.format(
                        sorted(list(attributes)),
                        sorted(list(keys)),
                        "/".join(map(str, path)),
                    ),
                )
            )

        return matches

    def check(self, properties, path, cfn):
        """Check itself"""
        matches = []

        property_sets = cfn.get_object_without_conditions(
            properties,
            [
                "AttributeDefinitions",
                "KeySchema",
                "GlobalSecondaryIndexes",
                "LocalSecondaryIndexes",
            ],
        )
        for property_set in property_sets:
            matches.extend(self.check_property_set(property_set, path))
        return matches

    def match_resource_properties(self, properties, _, path, cfn):
        """Match for sub properties"""
        matches = []
        matches.extend(self.check(properties, path, cfn))
        return matches



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from copy import deepcopy
from typing import Any, Dict, List, Union

import regex as re

import cfnlint.conditions
import cfnlint.helpers
from cfnlint.graph import Graph
from cfnlint.match import Match
from cfnlint.template.transforms import Transform

LOGGER = logging.getLogger(__name__)


def resolve_pointer(obj, pointer) -> Dict:
    """Find the elements at the end of a Cfn pointer

    Args:
        obj (dict): the root schema used for searching for the pointer
        pointer (str): the pointer using / to separate levels
    Returns:
        Dict: returns the object from the pointer
    """
    json_pointer = _SchemaPointer(obj, pointer)
    return json_pointer.resolve()


class _SchemaPointer:
    def __init__(self, obj: dict, pointer: str) -> None:
        self.obj = obj
        self.parts = pointer.split("/")[1:]

    def resolve(self) -> Dict:
        """Find the elements at the end of a Cfn pointer

        Args:
        Returns:
            Dict: returns the object from the pointer
        """
        obj = self.obj
        for part in self.parts:
            try:
                obj = self.walk(obj, part)
            except KeyError as e:
                raise e

        if "*" in self.parts:
            return {"type": "array", "items": obj}

        return obj

    # pylint: disable=too-many-return-statements
    def walk(self, obj: Dict, part: str) -> Any:
        """Walks one step in doc and returns the referenced part

        Args:
            obj (dict): the object to evaluate for the part
            part (str): the string representation of the part
        Returns:
            Dict: returns the object at the part
        """
        assert hasattr(obj, "__getitem__"), f"invalid document type {type(obj)}"

        try:
            # using a test for typeName as that is a root schema property
            if part == "properties" and obj.get("typeName"):
                return obj[part]
            if (
                obj.get("properties")
                and part != "definitions"
                and not obj.get("typeName")
            ):
                return obj["properties"][part]
            # arrays have a * in the path
            if part == "*" and obj.get("type") == "array":
                return obj.get("items")
            return obj[part]

        except KeyError as e:
            # CFN JSON pointers can go down $ref paths so lets do that if we can
            if obj.get("$ref"):
                try:
                    return resolve_pointer(self.obj, f"{obj.get('$ref')}/{part}")
                except KeyError as ke:
                    raise ke
            if obj.get("items", {}).get("$ref"):
                ref = obj.get("items", {}).get("$ref")
                try:
                    return resolve_pointer(self.obj, f"{ref}/{part}")
                except KeyError as ke:
                    raise ke
            if obj.get("oneOf"):
                for oneOf in obj.get("oneOf"):  # type: ignore
                    try:
                        return self.walk(oneOf, part)
                    except KeyError:
                        pass

                raise KeyError(f"No oneOf matches for {part}") from e
            raise e


class Template:  # pylint: disable=R0904,too-many-lines,too-many-instance-attributes
    """Class for a CloudFormation template"""

    # pylint: disable=dangerous-default-value
    def __init__(self, filename, template, regions=["us-east-1"]):
        self.filename = filename
        self.template = template
        self.regions = regions
        self.sections = [
            "AWSTemplateFormatVersion",
            "Description",
            "Metadata",
            "Parameters",
            "Mappings",
            "Conditions",
            "Transform",
            "Hooks",
            "Resources",
            "Outputs",
            "Rules",
        ]
        self.transform_pre = {}
        self.transform_pre["Globals"] = {}
        self.transform_pre["Ref"] = self.search_deep_keys("Ref")
        self.transform_pre["Fn::Sub"] = self.search_deep_keys("Fn::Sub")
        self.transform_pre["Fn::FindInMap"] = self.search_deep_keys("Fn::FindInMap")
        self.transform_pre["Transform"] = self.template.get("Transform", [])
        self.transform_pre["Fn::ForEach"] = self.search_deep_keys(
            cfnlint.helpers.FUNCTION_FOR_EACH
        )

        self.conditions = cfnlint.conditions.Conditions(self)
        self.__cache_search_deep_class = {}
        self.graph: Union[Graph, None] = None
        try:
            self.graph = Graph(self)
        except KeyError as err:
            LOGGER.debug(
                "Encountered KeyError error while building graph. Ignored as this "
                "should be caught by other rules and is more than likely a template "
                "formatting error: %s",
                err,
            )
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.info("Encountered unknown error while building graph: %s", err)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result

    def transform(self) -> List[Match]:
        """
        Transform the template
        """
        transform = Transform()
        return transform.transform(self)

    def build_graph(self):
        """Generates a DOT representation of the template"""
        path = self.filename + ".dot"
        try:
            self.graph.to_dot(path)
            LOGGER.info("DOT representation of the graph written to %s", path)
        except ImportError:
            LOGGER.error(
                "Could not write the graph in DOT format. Please install either `pygraphviz` or `pydot` modules."
            )

    def has_language_extensions_transform(self):
        """Check if the template has language extensions transform declared"""
        lang_extensions_transform = "AWS::LanguageExtensions"
        transform_declaration = self.transform_pre["Transform"]
        transform_type = (
            transform_declaration
            if isinstance(transform_declaration, list)
            else [transform_declaration]
        )
        return bool(lang_extensions_transform in transform_type)

    def has_serverless_transform(self):
        """Check if the template has SAM transform declared"""
        lang_extensions_transform = "AWS::Serverless-2016-10-31"
        transform_declaration = self.transform_pre["Transform"]
        transform_type = (
            transform_declaration
            if isinstance(transform_declaration, list)
            else [transform_declaration]
        )
        return bool(lang_extensions_transform in transform_type)

    def is_cdk_template(self) -> bool:
        """Check if the template was created by CDK"""
        resources = self.template.get("Resources")
        if not isinstance(resources, dict):
            return False

        for _, properties in resources.items():
            if not isinstance(properties, dict):
                continue
            resource_type = properties.get("Type")
            if not isinstance(resource_type, str):
                continue
            if resource_type == "AWS::CDK::Metadata":
                return True

        return False

    def get_resources(self, resource_type=[]):
        """
        Get Resources
        Filter on type when specified
        """
        LOGGER.debug("Get resources from template...")
        resources = self.template.get("Resources", {})
        if not isinstance(resources, dict):
            return {}
        if isinstance(resource_type, str):
            resource_type = [resource_type]

        results = {}
        for k, v in resources.items():
            if isinstance(v, dict):
                if (v.get("Type", None) in resource_type) or (
                    not resource_type and v.get("Type") is not None
                ):
                    results[k] = v

        return results

    def get_parameters(self):
        LOGGER.debug("Get parameters from template...")
        parameters = self.template.get("Parameters", {})
        if not parameters:
            return {}

        return parameters

    def get_parameters_valid(self):
        LOGGER.debug("Get parameters from template...")
        result = {}
        if isinstance(self.template.get("Parameters"), dict):
            parameters = self.template.get("Parameters")
            for parameter_name, parameter_value in parameters.items():
                if isinstance(parameter_value, dict):
                    if isinstance(parameter_value.get("Type"), str):
                        result[parameter_name] = parameter_value

        return result

    def get_outputs_valid(self):
        LOGGER.debug("Get outputs from template...")
        result = {}
        if isinstance(self.template.get("Outputs"), dict):
            parameters = self.template.get("Outputs")
            for parameter_name, parameter_value in parameters.items():
                if isinstance(parameter_value, dict):
                    if isinstance(parameter_value.get("Value"), (str, dict)):
                        result[parameter_name] = parameter_value

        return result

    def get_modules(self):
        """Get Modules"""
        LOGGER.debug("Get modules from template...")
        resources = self.template.get("Resources", {})
        if not resources:
            return {}

        results = {}
        for k, v in resources.items():
            if isinstance(v, dict):
                if v.get("Type") is not None and str(v.get("Type")).endswith(
                    "::MODULE"
                ):
                    results[k] = v

        return results

    def get_mappings(self):
        LOGGER.debug("Get mapping from template...")
        mappings = self.template.get("Mappings", {})
        if not mappings:
            return {}

        return mappings

    def get_resource_names(self):
        LOGGER.debug("Get the names of all resources from template...")
        results = []
        resources = self.template.get("Resources", {})
        if isinstance(resources, dict):
            for resourcename, _ in resources.items():
                results.append(resourcename)

        return results

    def get_parameter_names(self):
        LOGGER.debug("Get names of all parameters from template...")
        results = []
        parameters = self.template.get("Parameters", {})
        if isinstance(parameters, dict):
            for parametername, _ in parameters.items():
                results.append(parametername)

        return results

    def get_valid_refs(self):
        results = cfnlint.helpers.RegexDict()
        parameters = self.template.get("Parameters", {})
        if parameters:
            for name, value in parameters.items():
                if isinstance(value, dict):
                    if "Type" in value:
                        element = {}
                        element["Type"] = value["Type"]
                        element["From"] = "Parameters"
                        results[name] = element
        resources = self.template.get("Resources", {})
        if resources:
            for name, value in resources.items():
                resource_type = value.get("Type", "")
                if resource_type.endswith("::MODULE"):
                    element = {}
                    element["Type"] = "MODULE"
                    element["From"] = "Resources"
                    results[f"{name}.*"] = element
                elif resource_type:
                    element = {}
                    element["Type"] = resource_type
                    element["From"] = "Resources"
                    results[name] = element

        for pseudoparam in cfnlint.PSEUDOPARAMS:
            element = {}
            element["Type"] = "Pseudo"
            element["From"] = "Pseduo"
            results[pseudoparam] = element
        return results

    # pylint: disable=too-many-locals
    def get_valid_getatts(self):
        resourcetypes = cfnlint.helpers.RESOURCE_SPECS["us-east-1"].get("ResourceTypes")
        propertytypes = cfnlint.helpers.RESOURCE_SPECS["us-east-1"].get("PropertyTypes")
        results = {}
        resources = self.template.get("Resources", {})

        astrik_string_types = ("AWS::CloudFormation::Stack",)
        astrik_unknown_types = (
            "Custom::",
            "AWS::Serverless::",
            "AWS::CloudFormation::CustomResource",
        )

        def build_output_string(resource_type, property_name):
            prop = propertytypes.get(f"{resource_type}.{property_name}")
            if prop is None:
                yield None, None
            else:
                for k, v in prop.get("Properties", {}).items():
                    t = v.get("Type")
                    if t:
                        for item in build_output_string(resource_type, v):
                            yield f"{k}.{item[0]}", item[1]
                    else:
                        yield k, v.get("PrimitiveType")

        for name, value in resources.items():
            if "Type" in value:
                valtype = value["Type"]
                if isinstance(valtype, str):
                    if valtype.startswith(astrik_string_types):
                        LOGGER.debug(
                            "Cant build an appropriate getatt list from %s", valtype
                        )
                        results[name] = {"*": {"PrimitiveItemType": "String"}}
                    elif valtype.startswith(astrik_unknown_types) or valtype.endswith(
                        "::MODULE"
                    ):
                        LOGGER.debug(
                            "Cant build an appropriate getatt list from %s", valtype
                        )
                        results[name] = {"*": {}}
                    else:
                        if value["Type"] in resourcetypes:
                            if "Attributes" in resourcetypes[valtype]:
                                results[name] = {}
                                for attname, attvalue in resourcetypes[valtype][
                                    "Attributes"
                                ].items():
                                    if "Type" in attvalue:
                                        if attvalue.get("Type") in ["List", "Map"]:
                                            element = {}
                                            element.update(attvalue)
                                            results[name][attname] = element
                                        else:
                                            for item in build_output_string(
                                                value["Type"], attname
                                            ):
                                                if item[0] is None:
                                                    continue
                                                element = {"PrimitiveType": item[1]}
                                                results[name][
                                                    f"{attname}.{item[0]}"
                                                ] = element
                                    else:
                                        element = {}
                                        element.update(attvalue)
                                        results[name][attname] = element
                        for schema in cfnlint.helpers.REGISTRY_SCHEMAS:
                            if value["Type"] == schema["typeName"]:
                                results[name] = {}
                                for ro_property in schema["readOnlyProperties"]:
                                    try:
                                        item = resolve_pointer(schema, ro_property)
                                    except KeyError:
                                        continue
                                    item_type = item["type"]
                                    _type = None
                                    primitive_type = None
                                    if item_type == "string":
                                        primitive_type = "String"
                                    elif item_type == "number":
                                        primitive_type = "Double"
                                    elif item_type == "integer":
                                        primitive_type = "Integer"
                                    elif item_type == "boolean":
                                        primitive_type = "Boolean"
                                    elif item_type == "array":
                                        _type = "List"
                                        primitive_type = "String"

                                    ro_property = ro_property.replace(
                                        "/properties/", ""
                                    )
                                    results[name][".".join(ro_property.split("/"))] = {}
                                    if _type:
                                        results[name][".".join(ro_property.split("/"))][
                                            "Type"
                                        ] = _type
                                        results[name][".".join(ro_property.split("/"))][
                                            "PrimitiveItemType"
                                        ] = primitive_type
                                    elif primitive_type:
                                        results[name][".".join(ro_property.split("/"))][
                                            "PrimitiveType"
                                        ] = primitive_type

        return results

    def get_directives(self):
        results = {}
        resources = self.get_resources()
        if resources:
            for resource_name, resource_values in resources.items():
                if isinstance(resource_values, dict):
                    ignore_rule_ids = (
                        resource_values.get("Metadata", {})
                        .get("cfn-lint", {})
                        .get("config", {})
                        .get("ignore_checks", [])
                    )
                    for ignore_rule_id in ignore_rule_ids:
                        if ignore_rule_id not in results:
                            results[ignore_rule_id] = []
                        results[ignore_rule_id].append(resource_name)
        return results

    # pylint: disable=too-many-locals
    def _get_sub_resource_properties(self, keys, properties, path):
        """Used for recursive handling of properties in the keys"""
        LOGGER.debug("Get Sub Resource Properties from %s", keys)
        if not keys:
            result = {}
            result["Path"] = path
            result["Value"] = properties
            return [result]
        if isinstance(properties, dict):
            key = keys.pop(0)
            for key_name, key_value in properties.items():
                if key_name == key:
                    results = self._get_sub_resource_properties(
                        keys[:], key_value, path[:] + [key_name]
                    )
                    if results:
                        return results
        elif isinstance(properties, list):
            matches = []
            for index, item in enumerate(properties):
                results = None
                if isinstance(item, dict):
                    if len(item) == 1:
                        for sub_key, sub_value in item.items():
                            if sub_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                                cond_values = self.get_condition_values(sub_value)
                                results = []
                                for cond_value in cond_values:
                                    result_path = (
                                        path[:] + [index, sub_key] + cond_value["Path"]
                                    )
                                    results.extend(
                                        self._get_sub_resource_properties(
                                            keys[:], cond_value["Value"], result_path
                                        )
                                    )
                            elif sub_key == "Ref":
                                if sub_value != "AWS::NoValue":
                                    results = self._get_sub_resource_properties(
                                        keys[:], sub_value, path + [sub_key]
                                    )
                            else:
                                results = self._get_sub_resource_properties(
                                    keys[:], sub_value, path + [sub_key]
                                )
                    else:
                        results = self._get_sub_resource_properties(
                            keys[:], item, path + [index]
                        )
                if isinstance(results, dict):
                    matches.append(results)
                elif isinstance(results, list):
                    matches.extend(results)
            return matches

        return []

    def get_resource_properties(self, keys):
        """Filter keys of template"""
        LOGGER.debug("Get Properties from a resource: %s", keys)
        matches = []
        resourcetype = keys.pop(0)
        for resource_name, resource_value in self.get_resources(resourcetype).items():
            path = ["Resources", resource_name, "Properties"]
            properties = resource_value.get("Properties")
            if properties:
                results = self._get_sub_resource_properties(keys[:], properties, path)
                matches.extend(results)

        return matches

    # pylint: disable=dangerous-default-value
    def _search_deep_class(self, searchClass, cfndict, path):
        """Search deep for keys and get their values"""
        keys = []
        if isinstance(cfndict, searchClass):
            keys.append((path[:], cfndict))

        if isinstance(cfndict, dict):
            for key in cfndict:
                keys.extend(
                    self._search_deep_class(searchClass, cfndict[key], path[:] + [key])
                )
        elif isinstance(cfndict, list):
            for index, item in enumerate(cfndict):
                keys.extend(
                    self._search_deep_class(searchClass, item, path[:] + [index])
                )

        return keys

    def search_deep_class(self, searchClass, includeGlobals=True):
        """
        Search for a key in all parts of the template.
        :return if searchText is "Ref", an array like ['Resources', 'myInstance', 'Properties', 'ImageId', 'Ref', 'Ec2ImageId']
        """
        results = []
        if searchClass in self.__cache_search_deep_class:
            results = self.__cache_search_deep_class[searchClass]
        else:
            results.extend(self._search_deep_class(searchClass, self.template, []))
            self.__cache_search_deep_class[searchClass] = results

        # Globals are removed during a transform.  They need to be checked manually
        if includeGlobals:
            pre_results = self._search_deep_keys(
                searchClass, self.transform_pre.get("Globals"), []
            )
            for pre_result in pre_results:
                results.append(["Globals"] + pre_result)

        return results

    # pylint: disable=dangerous-default-value
    def _search_deep_keys(self, searchText: str | re.Pattern, cfndict, path):
        """Search deep for keys and get their values"""
        keys = []
        if isinstance(cfndict, dict):
            for key in cfndict:
                pathprop = path[:]
                pathprop.append(key)
                if isinstance(searchText, str):
                    if key == searchText:
                        pathprop.append(cfndict[key])
                        keys.append(pathprop)
                        # pop the last element off for nesting of found elements for
                        # dict and list checks
                        pathprop = pathprop[:-1]
                elif isinstance(searchText, re.Pattern):
                    if isinstance(key, str):
                        if re.match(searchText, key):
                            pathprop.append(cfndict[key])
                            keys.append(pathprop)
                            # pop the last element off for nesting of found elements for
                            # dict and list checks
                            pathprop = pathprop[:-1]
                if isinstance(cfndict[key], dict):
                    keys.extend(
                        self._search_deep_keys(searchText, cfndict[key], pathprop)
                    )
                elif isinstance(cfndict[key], list):
                    for index, item in enumerate(cfndict[key]):
                        pathproparr = pathprop[:]
                        pathproparr.append(index)
                        keys.extend(
                            self._search_deep_keys(searchText, item, pathproparr)
                        )
        elif isinstance(cfndict, list):
            for index, item in enumerate(cfndict):
                pathprop = path[:]
                pathprop.append(index)
                keys.extend(self._search_deep_keys(searchText, item, pathprop))

        return keys

    def search_deep_keys(
        self, searchText: str | re.Pattern, includeGlobals: bool = True
    ):
        """
        Search for a key in all parts of the template.
        :return if searchText is "Ref", an array like ['Resources', 'myInstance', 'Properties', 'ImageId', 'Ref', 'Ec2ImageId']
        """
        results = []
        results.extend(self._search_deep_keys(searchText, self.template, []))
        # Globals are removed during a transform.  They need to be checked manually
        if includeGlobals:
            pre_results = self._search_deep_keys(
                searchText, self.transform_pre.get("Globals"), []
            )
            for pre_result in pre_results:
                results.append(["Globals"] + pre_result)
        return results

    def get_condition_values(self, template, path=[]):
        """Evaluates conditions and brings back the values"""
        LOGGER.debug("Get condition values...")
        matches = []
        if not isinstance(template, list):
            return matches
        if not len(template) == 3:
            return matches

        for index, item in enumerate(template[1:]):
            result = {}
            result["Path"] = path[:] + [index + 1]
            if not isinstance(item, (dict, list)):
                # Just straight values and pass them through
                result["Value"] = item
                matches.append(result)
            elif len(item) == 1:
                # Checking for conditions inside of conditions
                if isinstance(item, dict):
                    for sub_key, sub_value in item.items():
                        if sub_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                            results = self.get_condition_values(
                                sub_value, result["Path"] + [sub_key]
                            )
                            if isinstance(results, list):
                                matches.extend(results)
                        elif sub_key == "Ref":
                            if sub_value != "AWS::NoValue":
                                result["Value"] = item
                                matches.append(result)
                        else:
                            # Return entire Item
                            result["Value"] = item
                            matches.append(result)
                else:
                    # Return entire Item
                    result["Value"] = item
                    matches.append(result)
            else:
                # Length longer than 1 means a list or object that should be fully returned
                result["Value"] = item
                matches.append(result)

        return matches

    def get_values(self, obj, key, path=[]):
        """
        Logic for getting the value of a key
        Returns None if the item isn't found
        Returns empty list if the item is found but Ref or GetAtt
        Returns all the values as a list if condition
        Returns the value if its just a string, int, boolean, etc.

        """
        LOGGER.debug("Get the value for key %s in %s", key, obj)
        matches = []

        if not isinstance(obj, dict):
            return None
        value = obj.get(key)
        if value is None:
            return None
        if isinstance(value, (dict)):
            if len(value) == 1:
                is_condition = False
                is_no_value = False
                for obj_key, obj_value in value.items():
                    if obj_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                        is_condition = True
                        results = self.get_condition_values(
                            obj_value, path[:] + [obj_key]
                        )
                        if isinstance(results, list):
                            for result in results:
                                check_obj = obj.copy()
                                check_obj[key] = result["Value"]
                                matches.extend(
                                    self.get_values(check_obj, key, result["Path"])
                                )
                    elif obj_key == "Ref" and obj_value == "AWS::NoValue":
                        is_no_value = True
                if not is_condition and not is_no_value:
                    result = {}
                    result["Path"] = path[:]
                    result["Value"] = value
                    matches.append(result)
            else:
                result = {}
                result["Path"] = path[:]
                result["Value"] = value
                matches.append(result)
        elif isinstance(value, (list)):
            for list_index, list_value in enumerate(value):
                if isinstance(list_value, dict):
                    if len(list_value) == 1:
                        is_condition = False
                        is_no_value = False
                        for obj_key, obj_value in list_value.items():
                            if obj_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                                is_condition = True
                                results = self.get_condition_values(
                                    obj_value, path[:] + [list_index, obj_key]
                                )
                                if isinstance(results, list):
                                    matches.extend(results)
                            elif obj_key == "Ref" and obj_value == "AWS::NoValue":
                                is_no_value = True
                        if not is_condition and not is_no_value:
                            result = {}
                            result["Path"] = path[:] + [list_index]
                            result["Value"] = list_value
                            matches.append(result)
                    else:
                        result = {}
                        result["Path"] = path[:] + [list_index]
                        result["Value"] = list_value
                        matches.append(result)
                else:
                    result = {}
                    result["Path"] = path[:] + [list_index]
                    result["Value"] = list_value
                    matches.append(result)
        else:
            result = {}
            result["Path"] = path[:]
            result["Value"] = value
            matches.append(result)

        return matches

    def _loc(self, obj):
        """Return location of object"""
        LOGGER.debug("Get location of object...")
        return (
            obj.start_mark.line,
            obj.start_mark.column,
            obj.end_mark.line,
            obj.end_mark.column,
        )

    def get_sub_parameters(self, sub_string):
        """Gets the parameters out of a Sub String"""
        results = []
        if not isinstance(sub_string, str):
            return results
        regex = re.compile(r"\${[^!].*?}")
        string_params = regex.findall(sub_string)

        for string_param in string_params:
            results.append(string_param[2:-1].strip())

        return results

    def get_location_yaml(self, text, path):
        """
        Get the location information
        """
        LOGGER.debug("Get location of path %s", path)
        result = None
        if not path:
            result = self._loc(text)
        elif len(path) > 1:
            try:
                result = self.get_location_yaml(text[path[0]], path[1:])
            except KeyError:
                pass
            # TypeError will help catch string indices must be integers for when
            # we parse JSON string and get a path inside that json string
            except TypeError:
                pass
            if not result:
                try:
                    for key in text:
                        if key == path[0]:
                            result = self._loc(key)
                except AttributeError as err:
                    LOGGER.debug(err)
        else:
            # If the last item of the path is an integer, and the vaue is an array,
            # Get the location of the item in the array
            if isinstance(text, list) and isinstance(path[0], int):
                try:
                    result = self._loc(text[path[0]])
                except AttributeError as err:
                    LOGGER.debug(err)
            else:
                try:
                    for key in text:
                        if key == path[0]:
                            result = self._loc(key)
                except AttributeError as err:
                    LOGGER.debug(err)

        return result

    def check_resource_property(
        self,
        resource_type,
        resource_property,
        check_value=None,
        check_ref=None,
        check_find_in_map=None,
        check_split=None,
        check_join=None,
        check_sub=None,
        **kwargs,
    ):
        """Check Resource Properties"""
        LOGGER.debug("Check property %s for %s", resource_property, resource_type)
        matches = []
        resources = self.get_resources(resource_type=resource_type)
        for resource_name, resource_object in resources.items():
            properties = resource_object.get("Properties", {})
            if properties:
                matches.extend(
                    self.check_value(
                        obj=properties,
                        key=resource_property,
                        path=["Resources", resource_name, "Properties"],
                        check_value=check_value,
                        check_ref=check_ref,
                        check_find_in_map=check_find_in_map,
                        check_split=check_split,
                        check_join=check_join,
                        check_sub=check_sub,
                        **kwargs,
                    )
                )
        return matches

    # pylint: disable=W0613,too-many-locals
    def check_value(
        self,
        obj,
        key,
        path,
        check_value=None,
        check_ref=None,
        check_get_att=None,
        check_find_in_map=None,
        check_split=None,
        check_join=None,
        check_import_value=None,
        check_sub=None,
        pass_if_null=False,
        **kwargs,
    ):
        LOGGER.debug("Check value %s for %s", key, obj)
        matches = []
        values_obj = self.get_values(obj=obj, key=key)
        new_path = path[:] + [key]
        if values_obj is None and pass_if_null:
            if check_value:
                matches.extend(
                    check_value(value=values_obj, path=new_path[:], **kwargs)
                )
        elif not values_obj:
            return matches
        else:
            for value_obj in values_obj:
                value = value_obj["Value"]
                child_path = value_obj["Path"]
                if not isinstance(value, dict):
                    if check_value:
                        matches.extend(
                            check_value(
                                value=value, path=new_path[:] + child_path, **kwargs
                            )
                        )
                else:
                    if len(value) == 1:
                        for dict_name, _ in value.items():
                            # If this is a function we shouldn't fall back to a check_value check
                            if dict_name in cfnlint.helpers.FUNCTIONS:
                                # convert the function name from camel case to underscore
                                # Example: Fn::FindInMap becomes check_find_in_map
                                function_name = f'check_{camel_to_snake(dict_name.replace("Fn::", ""))}'
                                if function_name == "check_ref":
                                    if check_ref:
                                        matches.extend(
                                            check_ref(
                                                value=value.get("Ref"),
                                                path=new_path[:] + child_path + ["Ref"],
                                                parameters=self.get_parameters(),
                                                resources=self.get_resources(),
                                                **kwargs,
                                            )
                                        )
                                else:
                                    if locals().get(function_name):
                                        matches.extend(
                                            locals()[function_name](
                                                value=value.get(dict_name),
                                                path=new_path[:]
                                                + child_path
                                                + [dict_name],
                                                **kwargs,
                                            )
                                        )
                            else:
                                if check_value:
                                    matches.extend(
                                        check_value(
                                            value=value,
                                            path=new_path[:] + child_path,
                                            **kwargs,
                                        )
                                    )
                    else:
                        if check_value:
                            matches.extend(
                                check_value(
                                    value=value, path=new_path[:] + child_path, **kwargs
                                )
                            )

        return matches

    def is_resource_available(self, path, resource):
        """
        Compares a path to resource to see if its available
        Returns scenarios that may result in the resource doesn't exist
        Input:
            Path: An array that is a Path to the object being checked
            Resource: The resource being compared to
        Output:
            If the resource is available the result is an empty array []
            If the resource is not available you will get a an array of Condition Names
                and when that condition is True or False will result in the resource
                not being available when trying to be associated.
                [{'ConditionName'}: False]
        """
        results = []
        path_conditions = self.get_conditions_from_path(self.template, path)
        resource_condition = (
            self.template.get("Resources", {}).get(resource, {}).get("Condition")
        )
        if isinstance(resource_condition, str):
            # if path conditions are empty that means its always true
            if not path_conditions:
                return [{resource_condition: False}]

            # resource conditions are always true.  If the same resource condition exists in the path
            # with the True then nothing else matters
            if True in path_conditions.get(resource_condition, {False}):
                return []

            # resource conditions are always true.  If the same resource condition exists in the path
            # with the False then nothing else matters
            if False in path_conditions.get(resource_condition, {True}):
                return [path_conditions]

            # if any condition paths loop back on themselves with the opposite then its unreachable code
            scenario = {}
            for condition_name, condition_bool in path_conditions.items():
                if len(condition_bool) > 1:
                    return results
                scenario[condition_name] = list(condition_bool)[0]

            if self.conditions.check_implies(scenario, resource_condition):
                return [{**{resource_condition: False}, **scenario}]

        # if resource condition isn't available then the resource is available
        return results

    def get_object_without_nested_conditions(self, obj, path, region=None):
        """
        Get a list of object values without conditions included.
        Evaluates deep into the object removing any nested conditions as well
        """
        results = []
        scenarios = self.get_condition_scenarios_below_path(path, False, region)
        if not isinstance(obj, (dict, list)):
            return results

        if not scenarios:
            if isinstance(obj, dict):
                if len(obj) == 1:
                    if obj.get("Ref") == "AWS::NoValue":
                        return results
            return [{"Scenario": None, "Object": obj}]

        def get_value(value, scenario):  # pylint: disable=R0911
            """Get the value based on the scenario resolving nesting"""
            if isinstance(value, dict):
                if len(value) == 1:
                    if "Fn::If" in value:
                        if_values = value.get("Fn::If")
                        if len(if_values) == 3:
                            if_path = scenario.get(if_values[0], None)
                            if if_path is not None:
                                if if_path:
                                    return get_value(if_values[1], scenario)
                                return get_value(if_values[2], scenario)
                    elif value.get("Ref") == "AWS::NoValue":
                        return None

                new_object = {}
                for k, v in value.items():
                    new_object[k] = get_value(v, scenario)
                return new_object
            if isinstance(value, list):
                new_list = []
                for item in value:
                    new_value = get_value(item, scenario)
                    if new_value is not None:
                        new_list.append(get_value(item, scenario))

                return new_list

            return value

        for scenario in scenarios:
            results.append({"Scenario": scenario, "Object": get_value(obj, scenario)})

        return results

    def get_value_from_scenario(self, obj, scenario):
        """
        Get object values from a provided scenario
        """

        def get_value(value, scenario):  # pylint: disable=R0911
            """Get the value based on the scenario resolving nesting"""
            if isinstance(value, dict):
                if len(value) == 1:
                    if "Fn::If" in value:
                        if_values = value.get("Fn::If")
                        if len(if_values) == 3:
                            if_path = scenario.get(if_values[0], None)
                            if if_path is not None:
                                if if_path:
                                    return get_value(if_values[1], scenario)
                                return get_value(if_values[2], scenario)
                    elif value.get("Ref") == "AWS::NoValue":
                        return None
                    else:
                        return value

                return value
            if isinstance(value, list):
                new_list = []
                for item in value:
                    new_value = get_value(item, scenario)
                    if new_value is not None:
                        new_list.append(get_value(item, scenario))

                return new_list

            return value

        result = cfnlint.dict_node({}, obj.start_mark, obj.end_mark)
        if isinstance(obj, dict):
            if len(obj) == 1:
                if obj.get("Fn::If"):
                    new_value = get_value(obj, scenario)
                    if new_value is not None:
                        result = new_value
                else:
                    for key, value in obj.items():
                        new_value = get_value(value, scenario)
                        if new_value is not None:
                            result[key] = new_value
            else:
                for key, value in obj.items():
                    new_value = get_value(value, scenario)
                    if new_value is not None:
                        result[key] = new_value

        return result

    def get_object_without_conditions(self, obj, property_names=None, region=None):
        """
        Gets a list of object values without conditions included
        Input:
            obj: The object/dict that makes up a set of properties
            Example:
                {
                    "DBSnapshotIdentifier" : {
                        "Fn::If" : [
                            "UseDBSnapshot",
                            {"Ref" : "DBSnapshotName"},
                            {"Ref" : "AWS::NoValue"}
                        ]
                    }
                }
        Output:
            A list of objects with scenarios for the conditions played out.
            If Ref to AWS::NoValue remove the property
            Example: [
                {
                    Object: {
                        "DBSnapshotIdentifier" : {"Ref" : "DBSnapshotName"}
                    },
                    Scenario: {UseDBSnapshot: True}
                }, {
                    Object: {
                    },
                    Scenario: {UseDBSnapshot: False}
                }
            ]
        """
        property_names = [] if property_names is None else property_names
        o = {}
        if property_names:
            for property_name in property_names:
                o[property_name] = deepcopy(obj.get(property_name))
        else:
            o = deepcopy(obj)
        results = []

        scenarios = self.get_conditions_scenarios_from_object([o], region)
        if not isinstance(obj, dict):
            return results

        if not scenarios:
            if isinstance(obj, dict):
                if len(obj) == 1:
                    if obj.get("Ref") == "AWS::NoValue":
                        return results
            return [{"Scenario": None, "Object": obj}]

        for scenario in scenarios:
            result_obj = self.get_value_from_scenario(obj, scenario)
            if result_obj:
                results.append({"Scenario": scenario, "Object": result_obj})

        return results

    def get_condition_scenarios_below_path(
        self, path, include_if_in_function=False, region=None
    ):
        """
        get Condition Scenarios from below path
        """
        fn_ifs = self.search_deep_keys("Fn::If")
        results = {}
        for fn_if in fn_ifs:
            if len(fn_if) >= len(path):
                if path == fn_if[0 : len(path)]:
                    # This needs to handle items only below the Path
                    result = self.get_conditions_from_path(
                        self.template, fn_if[0:-1], False, include_if_in_function
                    )
                    for condition_name, condition_values in result.items():
                        if condition_name in results:
                            results[condition_name].union(condition_values)
                        else:
                            results[condition_name] = condition_values

        return list(self.conditions.build_scenarios(list(results.keys()), region))

    def get_conditions_scenarios_from_object(self, objs, region=None):
        """
        Get condition from objects
        """

        def get_conditions_from_property(value):
            """Recursively get conditions"""
            results = set()
            if isinstance(value, dict):
                if len(value) == 1:
                    for k, v in value.items():
                        if k == "Fn::If":
                            if isinstance(v, list) and len(v) == 3:
                                if isinstance(v[0], str):
                                    results.add(v[0])
                                    results = results.union(
                                        get_conditions_from_property(v[1])
                                    )
                                    results = results.union(
                                        get_conditions_from_property(v[2])
                                    )
            elif isinstance(value, list):
                for v in value:
                    results = results.union(get_conditions_from_property(v))

            return results

        con = set()

        if isinstance(objs, dict):
            objs = [objs]

        for obj in objs:
            if isinstance(obj, dict):
                for k, v in obj.items():
                    # handle conditions directly under the object
                    if len(obj) == 1 and k == "Fn::If" and len(v) == 3:
                        con.add(v[0])
                        for r_c in v[1:]:
                            if isinstance(r_c, dict):
                                for s_k, s_v in r_c.items():
                                    if s_k == "Fn::If":
                                        con = con.union(
                                            get_conditions_from_property({s_k: s_v})
                                        )
                    else:
                        con = con.union(get_conditions_from_property(v))

        return list(self.conditions.build_scenarios(list(con), region))

    def get_conditions_from_path(
        self,
        text,
        path,
        include_resource_conditions=True,
        include_if_in_function=True,
        only_last=False,
    ):
        """
        Parent function to handle resources with conditions.
        Input:
            text: The object to start processing through the Path
            path: The path to recursively look for
        Output:
            An Object with keys being the Condition Names and the values are what
                if its in the True or False part of the path.
                {'condition': {True}}
        """

        results = self._get_conditions_from_path(
            text, path, include_if_in_function, only_last
        )
        if include_resource_conditions:
            if len(path) >= 2:
                if path[0] in ["Resources", "Outputs"]:
                    condition = text.get(path[0], {}).get(path[1], {}).get("Condition")
                    if isinstance(condition, str):
                        if not results.get(condition):
                            results[condition] = set()
                        results[condition].add(True)

        return results

    def _get_conditions_from_path(
        self, text, path, include_if_in_function=True, only_last=False
    ):
        """
        Get the conditions and their True/False value for the path provided
        Input:
            text: The object to start processing through the Path
            path: The path to recursively look for
        Output:
            An Object with keys being the Condition Names and the values are what
                if its in the True or False part of the path.
                {'condition': {True}}
        """
        LOGGER.debug("Get conditions for path %s", path)
        results = {}

        def get_condition_name(value, num=None):
            """Test conditions for validity before providing the name"""
            con_path = set()
            if num == 1:
                con_path.add(True)
            elif num == 2:
                con_path.add(False)
            else:
                con_path = con_path.union((True, False))

            if value:
                if isinstance(value, list):
                    if len(value) == 3:
                        if not results.get(value[0]):
                            results[value[0]] = set()
                        results[value[0]] = results[value[0]].union(con_path)

        try:
            # Found a condition at the root of the Path
            if path[0] == "Fn::If" and (
                (len(path) == 1 and only_last) or not only_last
            ):
                condition = text.get("Fn::If")
                if len(path) > 1:
                    if path[1] in [1, 2]:
                        get_condition_name(condition, path[1])
                else:
                    get_condition_name(condition)
            # Iterate if the Path has more than one value
            if len(path) > 1:
                if (
                    path[0] in cfnlint.helpers.FUNCTIONS and path[0] != "Fn::If"
                ) and not include_if_in_function:
                    return results
                child_results = self._get_conditions_from_path(
                    text[path[0]], path[1:], include_if_in_function, only_last
                )
                for c_r_k, c_r_v in child_results.items():
                    if not results.get(c_r_k):
                        results[c_r_k] = set()
                    results[c_r_k] = results[c_r_k].union(c_r_v)

        except KeyError as _:
            pass

        return results


def camel_to_snake(s):
    """
    Is it ironic that this function is written in camel case, yet it
    converts to snake case? hmm..
    """
    _underscorer1 = re.compile(r"(.)([A-Z][a-z]+)")
    _underscorer2 = re.compile("([a-z0-9])([A-Z])")
    subbed = _underscorer1.sub(r"\1_\2", s)
    return _underscorer2.sub(r"\1_\2", subbed).lower()



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class UpdateReplacePolicyDeletionPolicy(CloudFormationLintRule):
    """Check resources with UpdateReplacePolicy/DeletionPolicy have both"""

    id = "W3011"
    shortdesc = "Check resources with UpdateReplacePolicy/DeletionPolicy have both"
    description = "Both UpdateReplacePolicy and DeletionPolicy are needed to protect resources from deletion"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-deletionpolicy.html"
    tags = ["resources", "updatereplacepolicy", "deletionpolicy"]

    def match(self, cfn):
        """Check resources with UpdateReplacePolicy/DeletionPolicy have both"""
        matches = []

        for r_name, r_values in cfn.get_resources().items():
            if r_values.get("Type") not in [
                "AWS::Lambda::Version",
                "AWS::Lambda::LayerVersion",
            ]:
                # pylint: disable=too-many-boolean-expressions
                if (
                    r_values.get("DeletionPolicy")
                    and r_values.get("DeletionPolicy") != "Delete"
                    and not r_values.get("UpdateReplacePolicy")
                    or not r_values.get("DeletionPolicy")
                    and r_values.get("UpdateReplacePolicy")
                    and r_values.get("UpdateReplacePolicy") != "Delete"
                ):
                    path = ["Resources", r_name]
                    message = f'Both UpdateReplacePolicy and DeletionPolicy are needed to protect {"/".join(path)} from deletion'
                    matches.append(RuleMatch(path, message))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import logging
import os
import sys
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, Union

from jsonschema.exceptions import ValidationError

import cfnlint.config
import cfnlint.decode
import cfnlint.formatters
import cfnlint.maintenance
import cfnlint.runner
from cfnlint.helpers import REGIONS, REGISTRY_SCHEMAS
from cfnlint.rules import Match, ParseError, RulesCollection, TransformError
from cfnlint.template import Template

LOGGER = logging.getLogger("cfnlint")
DEFAULT_RULESDIR = os.path.join(os.path.dirname(__file__), "rules")
__CACHED_RULES = None

Matches = List[Match]
RulesCollectionNone = Optional[RulesCollection]
ArgsFilename = Tuple[
    cfnlint.config.ConfigMixIn, List[Optional[str]], cfnlint.formatters.BaseFormatter
]
TemplateRules = Tuple[Optional[str], RulesCollectionNone, Optional[Matches]]


class CfnLintExitException(Exception):
    """Generic exception used when the cli should exit"""

    def __init__(self, msg=None, exit_code=1):
        if msg is None:
            msg = f"process failed with exit code {exit_code}"
        super().__init__(msg)
        self.exit_code = exit_code


class InvalidRegionException(CfnLintExitException):
    """When an unsupported/invalid region is supplied"""


class UnexpectedRuleException(CfnLintExitException):
    """When processing a rule fails in an unexpected way"""


def run_cli(
    filename: str,
    template: str,
    rules: RulesCollectionNone,
    regions: Sequence[str],
    override_spec: dict,
    build_graph: bool,
    registry_schemas: Sequence[str],
    mandatory_rules: Optional[Sequence[str]] = None,
) -> Matches:
    """Process args and run"""

    if override_spec:
        cfnlint.helpers.override_specs(override_spec)

    if build_graph:
        template_obj = Template(filename, template, regions)
        template_obj.build_graph()

    if registry_schemas:
        for path in registry_schemas:
            if path and os.path.isdir(os.path.expanduser(path)):
                for f in os.listdir(path):
                    with open(os.path.join(path, f), encoding="utf-8") as schema:
                        REGISTRY_SCHEMAS.append(json.load(schema))

    return run_checks(filename, template, rules, regions, mandatory_rules)


def get_exit_code(matches: Matches, exit_level: str = "informational") -> int:
    """Determine exit code"""

    exit_levels: Dict[str, List[str]] = {
        "informational": ["informational", "warning", "error"],
        "warning": ["warning", "error"],
        "error": ["error"],
        "none": [],
    }

    exit_code = 0
    for match in matches:
        if (
            match.rule.severity == "informational"
            and match.rule.severity in exit_levels[exit_level]
        ):
            exit_code = exit_code | 8
        elif (
            match.rule.severity == "warning"
            and match.rule.severity in exit_levels[exit_level]
        ):
            exit_code = exit_code | 4
        elif (
            match.rule.severity == "error"
            and match.rule.severity in exit_levels[exit_level]
        ):
            exit_code = exit_code | 2

    return exit_code


# pylint: disable=too-many-return-statements


def get_formatter(fmt: str) -> cfnlint.formatters.BaseFormatter:
    if fmt:
        if fmt == "quiet":
            return cfnlint.formatters.QuietFormatter()
        if fmt == "parseable":
            # pylint: disable=bad-option-value
            return cfnlint.formatters.ParseableFormatter()
        if fmt == "json":
            return cfnlint.formatters.JsonFormatter()
        if fmt == "junit":
            return cfnlint.formatters.JUnitFormatter()
        if fmt == "pretty":
            return cfnlint.formatters.PrettyFormatter()
        if fmt == "sarif":
            return cfnlint.formatters.SARIFFormatter()

    return cfnlint.formatters.Formatter()


def get_rules(
    append_rules: List[str],
    ignore_rules: List[str],
    include_rules: List[str],
    configure_rules=None,
    include_experimental: bool = False,
    mandatory_rules: Union[List[str], None] = None,
    custom_rules: Union[str, None] = None,
) -> RulesCollection:
    rules = RulesCollection(
        ignore_rules,
        include_rules,
        configure_rules,
        include_experimental,
        mandatory_rules,
    )
    rules_paths: List[str] = [DEFAULT_RULESDIR] + append_rules
    try:
        for rules_path in rules_paths:
            if rules_path and os.path.isdir(os.path.expanduser(rules_path)):
                rules.create_from_directory(rules_path)
            else:
                rules.create_from_module(rules_path)

        rules.create_from_custom_rules_file(custom_rules)
    except (OSError, ImportError) as e:
        raise UnexpectedRuleException(
            f"Tried to append rules but got an error: {str(e)}", 1
        ) from e
    return rules


def get_matches(filenames: str, args: cfnlint.config.ConfigMixIn) -> Iterator[Match]:
    for filename in filenames:
        LOGGER.debug("Begin linting of file: %s", str(filename))
        (template, rules, errors) = get_template_rules(filename, args)
        # template matches may be empty but the template is still None
        # this happens when ignoring bad templates
        if not errors and template:
            matches = run_cli(
                filename,
                template,
                rules,
                args.regions,
                args.override_spec,
                args.build_graph,
                args.registry_schemas,
                args.mandatory_checks,
            )
            yield from iter(matches)
        else:
            if errors:
                yield from iter(errors)
        LOGGER.debug("Completed linting of file: %s", str(filename))


def configure_logging(debug_logging):
    """Backwards compatibility for integrators"""
    LOGGER.info(
        'Update your integrations to use "cfnlint.config.configure_logging" instead'
    )
    cfnlint.config.configure_logging(debug_logging, False)


def get_args_filenames(cli_args: Sequence[str]) -> ArgsFilename:
    """Get Template Configuration items and set them as default values"""
    try:
        config = cfnlint.config.ConfigMixIn(cli_args)
    except ValidationError as e:
        LOGGER.error("Error parsing config file: %s", str(e))
        sys.exit(1)

    fmt = config.format
    formatter = get_formatter(fmt)

    if config.update_specs:
        cfnlint.maintenance.update_resource_specs(config.force)
        sys.exit(0)

    if config.update_documentation:
        # Get ALL rules (ignore the CLI settings))
        documentation_rules = cfnlint.core.get_rules(
            [], [], ["I", "E", "W"], {}, True, []
        )
        cfnlint.maintenance.update_documentation(documentation_rules)
        sys.exit(0)

    if config.update_iam_policies:
        cfnlint.maintenance.update_iam_policies()
        sys.exit(0)

    if config.listrules:
        rules = cfnlint.core.get_rules(
            config.append_rules,
            config.ignore_checks,
            config.include_checks,
            config.configure_rules,
            config.mandatory_checks,
        )
        print(rules)
        sys.exit(0)

    if not sys.stdin.isatty() and not config.templates:
        return (config, [None], formatter)

    if not config.templates:
        # Not specified, print the help
        config.parser.print_help()
        sys.exit(1)

    return (config, config.templates, formatter)


def get_used_rules() -> Optional[RulesCollection]:
    return __CACHED_RULES


def _reset_rule_cache() -> None:
    """Reset the rule cache. Used mostly for testing"""
    global __CACHED_RULES  # pylint: disable=global-statement
    __CACHED_RULES = None


def _build_rule_cache(args: cfnlint.config.ConfigMixIn) -> None:
    global __CACHED_RULES  # pylint: disable=global-statement
    if __CACHED_RULES:
        __CACHED_RULES.configure(
            ignore_rules=args.ignore_checks,
            include_rules=args.include_checks,
            configure_rules=args.configure_rules,
            include_experimental=args.include_experimental,
            mandatory_rules=args.mandatory_checks,
        )
    else:
        __CACHED_RULES = cfnlint.core.get_rules(
            args.append_rules,
            args.ignore_checks,
            args.include_checks,
            args.configure_rules,
            args.include_experimental,
            args.mandatory_checks,
            args.custom_rules,
        )


def get_template_rules(
    filename: str, args: cfnlint.config.ConfigMixIn
) -> TemplateRules:
    """Get Template Configuration items and set them as default values"""

    ignore_bad_template: bool = False
    if args.ignore_bad_template:
        ignore_bad_template = True
    else:
        # There is no collection at this point so we need to handle this
        # check directly
        if not ParseError().is_enabled(
            include_experimental=False,
            ignore_rules=args.ignore_checks,
            include_rules=args.include_checks,
            mandatory_rules=args.mandatory_checks,
        ):
            ignore_bad_template = True

    (template, errors) = cfnlint.decode.decode(filename)

    if errors:
        _build_rule_cache(args)
        if len(errors) == 1 and ignore_bad_template and errors[0].rule.id == "E0000":
            return (template, __CACHED_RULES, [])
        return (template, __CACHED_RULES, errors)

    args.template_args = template

    _build_rule_cache(args)

    return (template, __CACHED_RULES, [])


def run_checks(
    filename: str,
    template: str,
    rules: RulesCollectionNone,
    regions: Sequence[str],
    mandatory_rules: Optional[Sequence[str]] = None,
) -> Matches:
    """Run Checks and Custom Rules against the template"""
    if regions:
        if not set(regions).issubset(set(REGIONS)):
            unsupported_regions = list(set(regions).difference(set(REGIONS)))
            msg = f"Regions {unsupported_regions} are unsupported. Supported regions are {REGIONS}"
            raise InvalidRegionException(msg, 32)

    errors: Matches = []

    if not isinstance(rules, RulesCollection):
        return []

    runner = cfnlint.runner.Runner(
        rules, filename, template, regions, mandatory_rules=mandatory_rules
    )

    # Transform logic helps with handling serverless templates
    ignore_transform_error: bool = False
    if isinstance(rules, RulesCollection) and not rules.is_rule_enabled(
        TransformError()
    ):
        ignore_transform_error = True

    errors.extend(runner.transform())

    if errors:
        if ignore_transform_error:
            return []  # if there is a transform error we can't continue

        return errors

    # Only do rule analysis if Transform was successful
    try:
        errors.extend(runner.run())
    except Exception as err:  # pylint: disable=W0703
        msg = f"Tried to process rules on file {filename} but got an error: {str(err)}"
        raise UnexpectedRuleException(msg, 1) from err
    errors.sort(key=lambda x: (x.filename, x.linenumber, x.rule.id))

    return errors



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class CircularDependency(CloudFormationLintRule):
    """Check if Resources have a circular dependency"""

    id = "E3004"
    shortdesc = "Resource dependencies are not circular"
    description = "Check that Resources are not circularly dependent by DependsOn, Ref, Sub, or GetAtt"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["resources", "circularly", "dependson", "ref", "sub", "getatt"]

    def match(self, cfn):
        matches = []

        if cfn.graph is None:
            return []
        for cycle in cfn.graph.get_cycles(cfn):
            source, target = cycle[:2]
            if (
                cfn.graph.graph.nodes[source].get("type") == "Resource"
                and cfn.graph.graph.nodes[target].get("type") == "Resource"
            ):
                message = f"Circular Dependencies for resource {source}. Circular dependency with [{target}]"
                path = ["Resources", source]
                matches.append(RuleMatch(path, message))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint.helpers import REGEX_DYN_REF, RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class StringSize(CloudFormationLintRule):
    """Check if a String has a length within the limit"""

    id = "E3033"
    shortdesc = "Check if a string has between min and max number of values specified"
    description = "Check strings for its length between the minimum and maximum"
    source_url = "https://github.com/awslabs/cfn-python-lint/blob/main/docs/cfn-resource-specification.md#allowedpattern"
    tags = ["resources", "property", "string", "size"]

    def initialize(self, cfn):
        """Initialize the rule"""
        for resource_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "ResourceTypes"
        ):
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "PropertyTypes"
        ):
            self.resource_sub_property_types.append(property_type_spec)

    def _check_string_length(self, value, path, **kwargs):
        """ """
        matches = []
        string_min = kwargs.get("string_min")
        string_max = kwargs.get("string_max")

        if isinstance(value, str) and not re.match(REGEX_DYN_REF, value):
            if not string_min <= len(value) <= string_max:
                message = "String has to have length between {0} and {1} at {2}. Current length {3}."
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            string_min, string_max, "/".join(map(str, path)), len(value)
                        ),
                    )
                )

        return matches

    def check(self, cfn, properties, specs, path):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for prop in p_value:
                if prop in specs:
                    value_type = specs.get(prop).get("Value", {}).get("ValueType", "")
                    if value_type:
                        if specs.get(prop).get("Type") == "List":
                            property_type = specs.get(prop).get("PrimitiveItemType")
                        else:
                            property_type = specs.get(prop).get("PrimitiveType")

                        value_specs = (
                            RESOURCE_SPECS.get(cfn.regions[0])
                            .get("ValueTypes")
                            .get(value_type, {})
                        )
                        if value_specs == "CACHED":
                            value_specs = (
                                RESOURCE_SPECS.get("us-east-1")
                                .get("ValueTypes")
                                .get(value_type, {})
                            )
                        if value_specs.get("StringMax") and value_specs.get(
                            "StringMin"
                        ):
                            if property_type == "String":
                                matches.extend(
                                    cfn.check_value(
                                        properties,
                                        prop,
                                        p_path,
                                        check_value=self._check_string_length,
                                        string_max=value_specs.get("StringMax"),
                                        string_min=value_specs.get("StringMin"),
                                    )
                                )
        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("PropertyTypes")
            .get(property_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("ResourceTypes")
            .get(resource_type, {})
            .get("Properties", {})
        )
        matches.extend(self.check(cfn, properties, specs, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Elb(CloudFormationLintRule):
    """Check if Elb Resource Properties"""

    id = "E2503"
    shortdesc = "Resource ELB Properties"
    description = "See if Elb Resource Properties are set correctly \
HTTPS has certificate HTTP has no certificate"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-elb-listener.html"
    tags = ["properties", "elb"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types = ["AWS::ElasticLoadBalancingV2::LoadBalancer"]

    def check_protocol_value(self, value, path, **kwargs):
        """
        Check Protocol Value
        """
        matches = []
        if isinstance(value, str):
            if value.upper() not in kwargs["accepted_protocols"]:
                message = "Protocol must be {0} is invalid at {1}"
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            (", ".join(kwargs["accepted_protocols"])),
                            ("/".join(map(str, path))),
                        ),
                    )
                )
            elif value.upper() in kwargs["certificate_protocols"]:
                if not kwargs["certificates"]:
                    message = (
                        "Certificates should be specified when using HTTPS for {0}"
                    )
                    matches.append(
                        RuleMatch(path, message.format(("/".join(map(str, path)))))
                    )

        return matches

    def get_loadbalancer_type(self, properties):
        """Check if type is application"""
        elb_type = properties.get("Type", "application")
        if isinstance(elb_type, str):
            if elb_type == "application":
                return "application"
            return "network"
        return None

    def check_alb_subnets(self, properties, path, scenario):
        """Validate at least two subnets with ALBs"""
        matches = []
        if self.get_loadbalancer_type(properties) == "application":
            subnets = properties.get("Subnets")
            if isinstance(subnets, list):
                if len(subnets) < 2:
                    if scenario:
                        message = 'You must specify at least two Subnets for load balancers with type "application" {0}'
                        scenario_text = " and ".join(
                            [
                                f'when condition "{k}" is {v}'
                                for (k, v) in scenario.items()
                            ]
                        )
                        matches.append(RuleMatch(path, message.format(scenario_text)))
                    else:
                        matches.append(
                            RuleMatch(
                                path[:] + ["Subnets"],
                                'You must specify at least two Subnets for load balancers with type "application"',
                            )
                        )
            subnet_mappings = properties.get("SubnetMappings")
            if isinstance(subnet_mappings, list):
                if len(subnet_mappings) < 2:
                    if scenario:
                        message = 'You must specify at least two SubnetMappings for load balancers with type "application" {0}'
                        scenario_text = " and ".join(
                            [
                                f'when condition "{k}" is {v}'
                                for (k, v) in scenario.items()
                            ]
                        )
                        matches.append(RuleMatch(path, message.format(scenario_text)))
                    else:
                        matches.append(
                            RuleMatch(
                                path[:] + ["SubnetMappings"],
                                'You must specify at least two SubnetMappings for load balancers with type "application"',
                            )
                        )

        return matches

    def match(self, cfn):
        """Check ELB Resource Parameters"""

        matches = []

        results = cfn.get_resource_properties(["AWS::ElasticLoadBalancingV2::Listener"])
        for result in results:
            matches.extend(
                cfn.check_value(
                    result["Value"],
                    "Protocol",
                    result["Path"],
                    check_value=self.check_protocol_value,
                    accepted_protocols=[
                        "GENEVE",
                        "HTTP",
                        "HTTPS",
                        "TCP",
                        "TCP_UDP",
                        "TLS",
                        "UDP",
                    ],
                    certificate_protocols=["HTTPS", "TLS"],
                    certificates=result["Value"].get("Certificates"),
                )
            )

        results = cfn.get_resource_properties(
            ["AWS::ElasticLoadBalancing::LoadBalancer", "Listeners"]
        )
        for result in results:
            if isinstance(result["Value"], list):
                for index, listener in enumerate(result["Value"]):
                    matches.extend(
                        cfn.check_value(
                            listener,
                            "Protocol",
                            result["Path"] + [index],
                            check_value=self.check_protocol_value,
                            accepted_protocols=["HTTP", "HTTPS", "TCP", "SSL"],
                            certificate_protocols=["HTTPS", "SSL"],
                            certificates=listener.get("SSLCertificateId"),
                        )
                    )

        return matches

    def match_resource_properties(self, resource_properties, _, path, cfn):
        """Check Load Balancers"""
        matches = []

        # Play out conditions to determine the relationship between property values
        scenarios = cfn.get_object_without_nested_conditions(resource_properties, path)
        for scenario in scenarios:
            properties = scenario.get("Object")

            matches.extend(
                self.check_alb_subnets(properties, path, scenario.get("Scenario"))
            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.helpers
from cfnlint.data import AdditionalSpecs
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class InstanceSize(CloudFormationLintRule):
    """Check if Resources RDS Instance Size is compatible with the RDS type"""

    id = "E3025"
    shortdesc = "RDS instance type is compatible with the RDS type"
    description = (
        "Check the RDS instance types are supported by the type of RDS engine. "
        "Only if the values are strings will this be checked."
    )
    source_url = "https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Concepts.DBInstanceClass.html"
    tags = ["resources", "rds"]

    valid_instance_types = cfnlint.helpers.load_resource(
        AdditionalSpecs, "RdsProperties.json"
    )

    def _get_license_model(self, engine, license_model):
        """Logic to get the correct license model"""
        if not license_model:
            if engine in self.valid_instance_types.get("license-included"):
                license_model = "license-included"
            elif engine in self.valid_instance_types.get("bring-your-own-license"):
                license_model = "bring-your-own-license"
            else:
                license_model = "general-public-license"
            self.logger.debug(
                "Based on Engine: %s we determined the default license will be %s",
                engine,
                license_model,
            )

        return license_model

    def get_resources(self, cfn):
        """Get resources that can be checked"""
        results = []
        for resource_name, resource_values in cfn.get_resources(
            "AWS::RDS::DBInstance"
        ).items():
            path = ["Resources", resource_name, "Properties"]
            properties = resource_values.get("Properties")
            # Properties items_safe heps remove conditions and focusing on the actual values and scenarios
            for prop_safe, prop_path_safe in properties.items_safe(path):
                engine = prop_safe.get("Engine")
                inst_class = prop_safe.get("DBInstanceClass")
                license_model = prop_safe.get("LicenseModel")

                # Need to get a default license model if none provided
                # Also need to validate all these values are strings otherwise we cannot
                # do validation
                if isinstance(engine, str) and isinstance(inst_class, str):
                    license_model = self._get_license_model(engine, license_model)
                    if isinstance(license_model, str):
                        results.append(
                            {
                                "Engine": engine,
                                "DBInstanceClass": inst_class,
                                "Path": prop_path_safe,
                                "LicenseModel": license_model,
                            }
                        )
                    else:
                        self.logger.debug(
                            "Skip evaluation based on [LicenseModel] not being a string."
                        )
                else:
                    self.logger.debug(
                        "Skip evaluation based on [Engine] or [DBInstanceClass] not being strings."
                    )

        return results

    def check_db_config(self, properties, region):
        """Check db properties"""
        matches = []

        db_engine = properties.get("Engine")
        db_license = properties.get("LicenseModel")
        db_instance_class = properties.get("DBInstanceClass")
        if db_license in self.valid_instance_types:
            if db_engine in self.valid_instance_types[db_license]:
                if region in self.valid_instance_types[db_license][db_engine]:
                    if (
                        db_instance_class
                        not in self.valid_instance_types[db_license][db_engine][region]
                    ):
                        message = 'DBInstanceClass "{0}" is not compatible with engine type "{1}" and LicenseModel "{2}" in region "{3}". Use instance types [{4}]'
                        matches.append(
                            RuleMatch(
                                properties.get("Path") + ["DBInstanceClass"],
                                message.format(
                                    db_instance_class,
                                    db_engine,
                                    db_license,
                                    region,
                                    ", ".join(
                                        map(
                                            str,
                                            self.valid_instance_types[db_license][
                                                db_engine
                                            ][region],
                                        )
                                    ),
                                ),
                            )
                        )
        else:
            self.logger.debug(
                "Skip evaluation based on license [%s] not matching.", db_license
            )
        return matches

    def match(self, cfn):
        """Check RDS Resource Instance Sizes"""

        matches = []
        resources = self.get_resources(cfn)
        for resource in resources:
            for region in cfn.regions:
                matches.extend(self.check_db_config(resource, region))
        return matches



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import itertools
import logging
import traceback
from typing import Any, Dict, Iterator, List, Tuple

from sympy import And, Implies, Not, Symbol
from sympy.assumptions.cnf import EncodedCNF
from sympy.logic.boolalg import BooleanFalse, BooleanTrue
from sympy.logic.inference import satisfiable

from cfnlint.conditions._utils import get_hash
from cfnlint.conditions.condition import ConditionNamed
from cfnlint.conditions.equals import Equal

LOGGER = logging.getLogger(__name__)


class Conditions:
    """Conditions provides the logic for relating individual condition together"""

    _conditions: Dict[str, ConditionNamed]
    _parameters: Dict[str, List[str]]  # Dict of parameters with AllowedValues hashed
    _max_scenarios: int = 128  # equivalent to 2^7

    def __init__(self, cfn):
        self._conditions = {}
        self._parameters = {}
        self._init_conditions(cfn=cfn)
        self._init_parameters(cfn=cfn)
        self._cnf, self._solver_params = self._build_cnf(list(self._conditions.keys()))

    def _init_conditions(self, cfn):
        conditions = cfn.template.get("Conditions")
        if isinstance(conditions, dict):
            for k, _ in conditions.items():
                try:
                    self._conditions[k] = ConditionNamed(k, conditions)
                except ValueError as e:
                    LOGGER.debug(
                        "Captured error while building condition %s: %s", k, str(e)
                    )
                except Exception as e:  # pylint: disable=broad-exception-caught
                    if LOGGER.getEffectiveLevel() == logging.DEBUG:
                        error_message = traceback.format_exc()
                    else:
                        error_message = str(e)
                    LOGGER.debug(
                        "Captured unknown error while building condition %s: %s",
                        k,
                        error_message,
                    )

    def _init_parameters(self, cfn: Any) -> None:
        parameters = cfn.template.get("Parameters")
        if not isinstance(parameters, dict):
            return
        for parameter_name, parameter in parameters.items():
            if not isinstance(parameter, dict):
                continue
            allowed_values = parameter.get("AllowedValues")
            if not allowed_values or not isinstance(allowed_values, list):
                continue

            param_hash = get_hash({"Ref": parameter_name})
            self._parameters[param_hash] = []
            for allowed_value in allowed_values:
                if isinstance(allowed_value, (str, int, float, bool)):
                    self._parameters[param_hash].append(get_hash(str(allowed_value)))

    def get(self, name: str, default: Any = None) -> ConditionNamed:
        """Return the conditions"""
        return self._conditions.get(name, default)

    def _build_cnf(
        self, condition_names: List[str]
    ) -> Tuple[EncodedCNF, Dict[str, Any]]:
        cnf = EncodedCNF()

        # build parameters and equals into solver
        equal_vars: Dict[str, Symbol] = {}

        equals: Dict[str, Equal] = {}
        for condition_name in condition_names:
            c_equals = self._conditions[condition_name].equals
            for c_equal in c_equals:
                # check to see if equals already matches another one
                if c_equal.hash in equal_vars:
                    continue

                if c_equal.is_static is not None:
                    if c_equal.is_static:
                        equal_vars[c_equal.hash] = BooleanTrue()
                    else:
                        equal_vars[c_equal.hash] = BooleanFalse()
                else:
                    equal_vars[c_equal.hash] = Symbol(c_equal.hash)
                    # See if parameter in this equals is the same as another equals
                    for param in c_equal.parameters:
                        for e_hash, e_equals in equals.items():
                            if param in e_equals.parameters:
                                # equivalent to NAND logic. We want to make sure that both equals
                                # are not both True at the same time
                                cnf.add_prop(
                                    ~(equal_vars[c_equal.hash] & equal_vars[e_hash])
                                )
                equals[c_equal.hash] = c_equal

        # Determine if a set of conditions can never be all false
        allowed_values = self._parameters.copy()
        if allowed_values:
            # iteration 1 cleans up all the hash values from allowed_values to know if we
            # used them all
            for _, equal_1 in equals.items():
                for param in equal_1.parameters:
                    if param.hash not in allowed_values:
                        continue
                    if isinstance(equal_1.left, str):
                        if get_hash(equal_1.left) in allowed_values[param.hash]:
                            allowed_values[param.hash].remove(get_hash(equal_1.left))
                        else:
                            equal_vars[equal_1.hash] = BooleanFalse()
                    if isinstance(equal_1.right, str):
                        if get_hash(equal_1.right) in allowed_values[param.hash]:
                            allowed_values[param.hash].remove(get_hash(equal_1.right))
                        else:
                            equal_vars[equal_1.hash] = BooleanFalse()

            # iteration 2 builds the cnf formulas to make sure any empty lists
            # are now full not equals
            for allowed_hash, allowed_value in allowed_values.items():
                # means the list is empty and all allowed values are validated
                # so not all equals can be false
                if not allowed_value:
                    prop = None
                    for _, equal_1 in equals.items():
                        for param in equal_1.parameters:
                            if allowed_hash == param.hash:
                                if prop is None:
                                    prop = Not(equal_vars[equal_1.hash])
                                else:
                                    prop = prop & Not(equal_vars[equal_1.hash])
                    # Need to make sure they aren't all False
                    # So Not(Not(Equal1) & Not(Equal2))
                    # When Equal1 False and Equal2 False
                    # Not(True & True) = False allowing this not to happen
                    if prop is not None:
                        cnf.add_prop(Not(prop))

        return (cnf, equal_vars)

    def build_scenarios(
        self, condition_names: List[str], region: None = None
    ) -> Iterator[Dict[str, bool]]:
        """Given a list of condition names this function will yield scenarios that represent
        those conditions and there result (True/False)

        Args:
            condition_names (List[str]): A list of condition names

        Returns:
            Iterator[Dict[str, bool]]: yield dict objects of {ConditionName: True/False}
        """
        # nothing to yield if there are no conditions
        if len(condition_names) == 0:
            return

        try:
            # build a large matric of True/False options based on the provided conditions
            scenarios_returned = 0
            if region:
                products = itertools.starmap(
                    self.build_scenerios_on_region,
                    itertools.product(condition_names, [region]),
                )
            else:
                products = itertools.product([True, False], repeat=len(condition_names))
            for p in products:
                cnf = self._cnf.copy()
                params = dict(zip(condition_names, p))
                for condition_name, opt in params.items():
                    if opt:
                        cnf.add_prop(
                            self._conditions[condition_name].build_true_cnf(
                                self._solver_params
                            )
                        )
                    else:
                        cnf.add_prop(
                            self._conditions[condition_name].build_false_cnf(
                                self._solver_params
                            )
                        )

                # if the scenario can be satisfied then return it
                if satisfiable(cnf):
                    yield params
                    scenarios_returned += 1

                # On occassions people will use a lot of non-related conditions
                # this is fail safe to limit the maximum number of responses
                if scenarios_returned >= self._max_scenarios:
                    return
        except KeyError:
            # KeyError is because the listed condition doesn't exist because of bad
            #  formatting or just the wrong condition name
            return

    def check_implies(self, scenarios: Dict[str, bool], implies: str) -> bool:
        """Based on a bunch of scenario conditions and their Truth/False value
        determine if implies condition is True any time the scenarios are satisfied
        solver, solver_params = self._build_solver(list(scenarios.keys()) + [implies])

        Args:
            scenarios (Dict[str, bool]): A list of condition names and if they are True or False
            implies: the condition name that we are implying will also be True

        Returns:
            bool: if the implied condition will be True if the scenario is True
        """
        try:
            cnf = self._cnf.copy()
            # if the implies condition has to be false in the scenarios we
            # know it can never be true
            if not scenarios.get(implies, True):
                return False

            conditions = []
            for condition_name, opt in scenarios.items():
                if opt:
                    conditions.append(
                        self._conditions[condition_name].build_true_cnf(
                            self._solver_params
                        )
                    )
                else:
                    conditions.append(
                        self._conditions[condition_name].build_false_cnf(
                            self._solver_params
                        )
                    )

            implies_condition = self._conditions[implies].build_true_cnf(
                self._solver_params
            )

            and_condition = And(*conditions)
            cnf.add_prop(and_condition)

            # if the implies condition has to be true already then we don't
            # need to imply it
            if not scenarios.get(implies):
                cnf.add_prop(Not(Implies(and_condition, implies_condition)))
            if satisfiable(cnf):
                return True

            return False
        except KeyError:
            # KeyError is because the listed condition doesn't exist because of bad
            #  formatting or just the wrong condition name
            return True

    def build_scenerios_on_region(
        self, condition_name: str, region: str
    ) -> Iterator[bool]:
        """Based on a region validate if the condition_name could be true

        Args:
            condition_name (str): The name of the condition we are validating against
            region (str): the name of the region

        Returns:
            Generator[bool]: Returns True, False, or True and False depending on if the
               condition could be True, False or both based on the region parameter
        """
        if not isinstance(condition_name, str):
            return
        cnf_region = self._cnf.copy()
        found_region = False
        if condition_name not in self._conditions:
            return
        for eql in self._conditions[condition_name].equals:
            is_region, equal_region = eql.is_region
            if is_region:
                found_region = True
                if equal_region == region:
                    cnf_region.add_prop(And(self._solver_params[eql.hash]))
                else:
                    cnf_region.add_prop(Not(self._solver_params[eql.hash]))

        # The condition doesn't use a region parameter so it can be True or False
        # Note: It is possible its a hard coded condition but
        # for now we will return True and False
        if not found_region:
            yield True
            yield False
            return

        cnf_test = cnf_region.copy()
        cnf_test.add_prop(
            self._conditions[condition_name].build_true_cnf(self._solver_params)
        )
        if satisfiable(cnf_test):
            yield True

        cnf_test = cnf_region.copy()
        cnf_test.add_prop(
            self._conditions[condition_name].build_false_cnf(self._solver_params)
        )
        if satisfiable(cnf_test):
            yield False



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import regex as re

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class RefExist(CloudFormationLintRule):
    id = "E1012"
    shortdesc = "Check if Refs exist"
    description = "Making sure the refs exist"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-ref.html"
    tags = ["functions", "ref"]

    def searchstring(self, string):
        """Search string for tokenized fields"""
        regex = re.compile(r"\${([a-zA-Z0-9]*)}")
        return regex.findall(string)

    def match(self, cfn):
        matches = []

        # Build the list of refs
        reftrees = cfn.search_deep_keys("Ref")
        refs = []
        for reftree in reftrees:
            refs.append(reftree[-1])

        valid_refs = cfn.get_valid_refs()

        # start with the basic ref calls
        for reftree in reftrees:
            ref = reftree[-1]
            if isinstance(ref, (str, int)):
                if ref not in valid_refs:
                    message = "Ref {0} not found as a resource or parameter"
                    matches.append(RuleMatch(reftree[:-2], message.format(ref)))

        return matches
